# frozen_string_literal: true

require_relative "module/aliasing"
require_relative "module/introspection"
require_relative "module/anonymous"
require_relative "module/attribute_accessors"
require_relative "module/attribute_accessors_per_thread"
require_relative "module/attr_internal"
require_relative "module/concerning"
require_relative "module/delegation"
require_relative "module/deprecation"
require_relative "module/redefine_method"
require_relative "module/remove_method"
